@@ -4,9 +4,11 @@
 Scapy *BSD native support - BPF sockets
 """
 
+from ctypes import c_long, sizeof
 import errno
 import fcntl
 import os
+import platform
 from select import select
 import struct
 import time
@@ -19,11 +21,16 @@ from scapy.config import conf
 from scapy.consts import FREEBSD, NETBSD, DARWIN
 from scapy.data import ETH_P_ALL
 from scapy.error import Scapy_Exception, warning
+from scapy.interfaces import network_name
 from scapy.supersocket import SuperSocket
 from scapy.compat import raw
+from scapy.layers.l2 import Loopback
 
 
-if FREEBSD or NETBSD:
+if FREEBSD:
+    # On 32bit architectures long might be 32bit.
+    BPF_ALIGNMENT = sizeof(c_long)
+elif NETBSD:
     BPF_ALIGNMENT = 8  # sizeof(long)
 else:
     BPF_ALIGNMENT = 4  # sizeof(int32_t)
@@ -48,12 +55,10 @@ class _L2bpfSocket(SuperSocket):
         else:
             self.promisc = promisc
 
-        if iface is None:
-            self.iface = conf.iface
-        else:
-            self.iface = iface
+        self.iface = network_name(iface or conf.iface)
 
         # Get the BPF handle
+        self.ins = None
         (self.ins, self.dev_bpf) = get_dev_bpf()
         self.outs = self.ins
 
@@ -102,6 +107,7 @@ class _L2bpfSocket(SuperSocket):
                                   self.dev_bpf)
 
         # Configure the BPF filter
+        filter_attached = False
         if not nofilter:
             if conf.except_filter:
                 if filter:
@@ -109,7 +115,21 @@ class _L2bpfSocket(SuperSocket):
                 else:
                     filter = "not (%s)" % conf.except_filter
             if filter is not None:
+                try:
+                    attach_filter(self.ins, filter, self.iface)
+                    filter_attached = True
+                except ImportError as ex:
+                    warning("Cannot set filter: %s" % ex)
+        if NETBSD and filter_attached is False:
+            # On NetBSD, a filter must be attached to an interface, otherwise
+            # no frame will be received by os.read(). When no filter has been
+            # configured, Scapy uses a simple tcpdump filter that does nothing
+            # more than ensuring the length frame is not null.
+            filter = "greater 0"
+            try:
                 attach_filter(self.ins, filter, self.iface)
+            except ImportError as ex:
+                warning("Cannot set filter: %s" % ex)
 
         # Set the guessed packet class
         self.guessed_cls = self.guess_cls()
@@ -223,7 +243,7 @@ class _L2bpfSocket(SuperSocket):
         the available sockets.
         """
         # sockets, None (means use the socket's recv() )
-        return bpf_select(sockets, remain), None
+        return bpf_select(sockets, remain)
 
 
 class L2bpfListenSocket(_L2bpfSocket):
@@ -260,8 +280,21 @@ class L2bpfListenSocket(_L2bpfSocket):
             return
 
         # Extract useful information from the BPF header
-        if FREEBSD or NETBSD:
-            # struct bpf_xhdr or struct bpf_hdr32
+        if FREEBSD:
+            # Unless we set BIOCSTSTAMP to something different than
+            # BPF_T_MICROTIME, we will get bpf_hdr on FreeBSD, which means
+            # that we'll get a struct timeval, which is time_t, suseconds_t.
+            # On i386 time_t is 32bit so the bh_tstamp will only be 8 bytes.
+            # We really want to set BIOCSTSTAMP to BPF_T_NANOTIME and be
+            # done with this and it always be 16?
+            if platform.machine() == "i386":
+                # struct bpf_hdr
+                bh_tstamp_offset = 8
+            else:
+                # struct bpf_hdr (64bit time_t) or struct bpf_xhdr
+                bh_tstamp_offset = 16
+        elif NETBSD:
+            # struct bpf_hdr or struct bpf_hdr32
             bh_tstamp_offset = 16
         else:
             # struct bpf_hdr
@@ -336,7 +369,9 @@ class L3bpfSocket(L2bpfSocket):
         """Receive on layer 3"""
         r = SuperSocket.recv(self, x)
         if r:
+            r.payload.time = r.time
             return r.payload
+        return r
 
     def send(self, pkt):
         """Send a packet"""
@@ -355,7 +390,18 @@ class L3bpfSocket(L2bpfSocket):
             self.assigned_interface = iff
 
         # Build the frame
-        frame = raw(self.guessed_cls() / pkt)
+        if self.guessed_cls == Loopback:
+            # bpf(4) man page (from macOS, but also for BSD):
+            # "A packet can be sent out on the network by writing to a bpf
+            # file descriptor. [...] Currently only writes to Ethernets and
+            # SLIP links are supported"
+            #
+            # Headers are only mentioned for reads, not writes. tuntaposx's tun
+            # device reports as a "loopback" device, but it does IP.
+            frame = raw(pkt)
+        else:
+            frame = raw(self.guessed_cls() / pkt)
+
         pkt.sent_time = time.time()
 
         # Send the frame
