@@ -6,7 +6,3 @@
 """
 Package holding ASN.1 related modules.
 """
-
-# We do not import mib.py because it is more bound to scapy and
-# less prone to be used in a standalone fashion
-__all__ = ["asn1", "ber"]
