@@ -1,5 +1,3 @@
-#! /usr/bin/env python
-
 # This file is part of Scapy
 # See http://www.secdev.org/projects/scapy for more information
 # Copyright (C) Nils Weiss <nils@we155.de>
@@ -13,29 +11,40 @@ import struct
 from scapy.fields import ObservableDict, XByteEnumField, ByteEnumField, \
     ConditionalField, XByteField, StrField, XShortEnumField, XShortField, \
     X3BytesField, XIntField, ShortField, PacketField, PacketListField, \
-    FieldListField
-from scapy.packet import Packet, bind_layers
+    FieldListField, MultipleTypeField, StrFixedLenField
+from scapy.packet import Packet, bind_layers, NoPayload
 from scapy.config import conf
-from scapy.error import warning
-from scapy.utils import PeriodicSenderThread
+from scapy.error import warning, log_loading
+from scapy.contrib.isotp import ISOTP
 
 
 """
 GMLAN
 """
 
-conf.contribs['GMLAN'] = {'GMLAN_ECU_AddressingScheme': None}
+try:
+    if conf.contribs['GMLAN']['treat-response-pending-as-answer']:
+        pass
+except KeyError:
+    log_loading.info("Specify \"conf.contribs['GMLAN'] = "
+                     "{'treat-response-pending-as-answer': True}\" to treat "
+                     "a negative response 'RequestCorrectlyReceived-"
+                     "ResponsePending' as answer of a request. \n"
+                     "The default value is False.")
+    conf.contribs['GMLAN'] = {'treat-response-pending-as-answer': False}
+
+conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'] = None
 
 
-class GMLAN(Packet):
+class GMLAN(ISOTP):
     @staticmethod
     def determine_len(x):
         if conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'] is None:
-            warning("Define conf.GMLAN_ECU_AddressingScheme! "
+            warning("Define conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme']! "  # noqa: E501
                     "Assign either 2,3 or 4")
         if conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'] \
                 not in [2, 3, 4]:
-            warning("Define conf.GMLAN_ECU_AddressingScheme! "
+            warning("Define conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme']! "  # noqa: E501
                     "Assign either 2,3 or 4")
         return conf.contribs['GMLAN']['GMLAN_ECU_AddressingScheme'] == x
 
@@ -87,12 +96,17 @@ class GMLAN(Packet):
     ]
 
     def answers(self, other):
-        """DEV: true if self is an answer from other"""
-        if other.__class__ == self.__class__:
-            return (other.service + 0x40) == self.service or \
-                   (self.service == 0x7f and
-                    (self.requestServiceId == other.service))
-        return 0
+        if not isinstance(other, type(self)):
+            return False
+        if self.service == 0x7f:
+            return self.payload.answers(other)
+        if self.service == (other.service + 0x40):
+            if isinstance(self.payload, NoPayload) or \
+                    isinstance(other.payload, NoPayload):
+                return True
+            else:
+                return self.payload.answers(other.payload)
+        return False
 
     def hashret(self):
         if self.service == 0x7f:
@@ -150,6 +164,10 @@ class GMLAN_RFRDPR(Packet):
         ByteEnumField('subfunction', 0, GMLAN_RFRD.subfunctions)
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_RFRD) and \
+            other.subfunction == self.subfunction
+
 
 bind_layers(GMLAN, GMLAN_RFRDPR, service=0x52)
 
@@ -282,6 +300,10 @@ class GMLAN_RDBIPR(Packet):
         XByteEnumField('dataIdentifier', 0, GMLAN_RDBI.dataIdentifiers),
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_RDBI) and \
+            other.dataIdentifier == self.dataIdentifier
+
 
 bind_layers(GMLAN, GMLAN_RDBIPR, service=0x5A)
 
@@ -310,6 +332,10 @@ class GMLAN_RDBPIPR(Packet):
         XShortEnumField('parameterIdentifier', 0, GMLAN_RDBPI.dataIdentifiers),
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_RDBPI) and \
+            self.parameterIdentifier in other.identifiers
+
 
 bind_layers(GMLAN, GMLAN_RDBPIPR, service=0x62)
 
@@ -340,12 +366,16 @@ bind_layers(GMLAN, GMLAN_RDBPKTI, service=0xAA)
 class GMLAN_RMBA(Packet):
     name = 'ReadMemoryByAddress'
     fields_desc = [
-        ConditionalField(XShortField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(2)),
-        ConditionalField(X3BytesField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(3)),
-        ConditionalField(XIntField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(4)),
+        MultipleTypeField(
+            [
+                (XShortField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(2)),
+                (X3BytesField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(3)),
+                (XIntField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(4))
+            ],
+            XIntField('memoryAddress', 0)),
         XShortField('memorySize', 0),
     ]
 
@@ -356,15 +386,23 @@ bind_layers(GMLAN, GMLAN_RMBA, service=0x23)
 class GMLAN_RMBAPR(Packet):
     name = 'ReadMemoryByAddressPositiveResponse'
     fields_desc = [
-        ConditionalField(XShortField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(2)),
-        ConditionalField(X3BytesField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(3)),
-        ConditionalField(XIntField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(4)),
-        StrField('dataRecord', None, fmt="B")
+        MultipleTypeField(
+            [
+                (XShortField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(2)),
+                (X3BytesField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(3)),
+                (XIntField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(4))
+            ],
+            XIntField('memoryAddress', 0)),
+        StrField('dataRecord', b"", fmt="B")
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_RMBA) and \
+            other.memoryAddress == self.memoryAddress
+
 
 bind_layers(GMLAN, GMLAN_RMBAPR, service=0x63)
 
@@ -389,7 +427,7 @@ class GMLAN_SA(Packet):
     name = 'SecurityAccess'
     fields_desc = [
         ByteEnumField('subfunction', 0, subfunctions),
-        ConditionalField(XShortField('securityKey', B""),
+        ConditionalField(XShortField('securityKey', 0),
                          lambda pkt: pkt.subfunction % 2 == 0)
     ]
 
@@ -401,10 +439,14 @@ class GMLAN_SAPR(Packet):
     name = 'SecurityAccessPositiveResponse'
     fields_desc = [
         ByteEnumField('subfunction', 0, GMLAN_SA.subfunctions),
-        ConditionalField(XShortField('securitySeed', B""),
+        ConditionalField(XShortField('securitySeed', 0),
                          lambda pkt: pkt.subfunction % 2 == 1),
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_SA) \
+            and other.subfunction == self.subfunction
+
 
 bind_layers(GMLAN, GMLAN_SAPR, service=0x67)
 
@@ -427,6 +469,10 @@ class GMLAN_DDMPR(Packet):
         XByteField('DPIDIdentifier', 0)
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_DDM) \
+            and other.DPIDIdentifier == self.DPIDIdentifier
+
 
 bind_layers(GMLAN, GMLAN_DDMPR, service=0x6C)
 
@@ -436,12 +482,16 @@ class GMLAN_DPBA(Packet):
     name = 'DefinePIDByAddress'
     fields_desc = [
         XShortField('parameterIdentifier', 0),
-        ConditionalField(XShortField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(2)),
-        ConditionalField(X3BytesField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(3)),
-        ConditionalField(XIntField('memoryAddress', 0),
-                         lambda pkt: GMLAN.determine_len(4)),
+        MultipleTypeField(
+            [
+                (XShortField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(2)),
+                (X3BytesField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(3)),
+                (XIntField('memoryAddress', 0),
+                 lambda pkt: GMLAN.determine_len(4))
+            ],
+            XIntField('memoryAddress', 0)),
         XByteField('memorySize', 0),
     ]
 
@@ -455,6 +505,10 @@ class GMLAN_DPBAPR(Packet):
         XShortField('parameterIdentifier', 0),
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_DPBA) \
+            and other.parameterIdentifier == self.parameterIdentifier
+
 
 bind_layers(GMLAN, GMLAN_DPBA, service=0x6D)
 
@@ -464,12 +518,16 @@ class GMLAN_RD(Packet):
     name = 'RequestDownload'
     fields_desc = [
         XByteField('dataFormatIdentifier', 0),
-        ConditionalField(XShortField('memorySize', 0),
-                         lambda pkt: GMLAN.determine_len(2)),
-        ConditionalField(X3BytesField('memorySize', 0),
-                         lambda pkt: GMLAN.determine_len(3)),
-        ConditionalField(XIntField('memorySize', 0),
-                         lambda pkt: GMLAN.determine_len(4)),
+        MultipleTypeField(
+            [
+                (XShortField('memorySize', 0),
+                 lambda pkt: GMLAN.determine_len(2)),
+                (X3BytesField('memorySize', 0),
+                 lambda pkt: GMLAN.determine_len(3)),
+                (XIntField('memorySize', 0),
+                 lambda pkt: GMLAN.determine_len(4))
+            ],
+            XIntField('memorySize', 0))
     ]
 
 
@@ -485,13 +543,17 @@ class GMLAN_TD(Packet):
     name = 'TransferData'
     fields_desc = [
         ByteEnumField('subfunction', 0, subfunctions),
-        ConditionalField(XShortField('startingAddress', 0),
-                         lambda pkt: GMLAN.determine_len(2)),
-        ConditionalField(X3BytesField('startingAddress', 0),
-                         lambda pkt: GMLAN.determine_len(3)),
-        ConditionalField(XIntField('startingAddress', 0),
-                         lambda pkt: GMLAN.determine_len(4)),
-        StrField("dataRecord", None)
+        MultipleTypeField(
+            [
+                (XShortField('startingAddress', 0),
+                 lambda pkt: GMLAN.determine_len(2)),
+                (X3BytesField('startingAddress', 0),
+                 lambda pkt: GMLAN.determine_len(3)),
+                (XIntField('startingAddress', 0),
+                 lambda pkt: GMLAN.determine_len(4))
+            ],
+            XIntField('startingAddress', 0)),
+        StrField("dataRecord", b"")
     ]
 
 
@@ -503,7 +565,7 @@ class GMLAN_WDBI(Packet):
     name = 'WriteDataByIdentifier'
     fields_desc = [
         XByteEnumField('dataIdentifier', 0, GMLAN_RDBI.dataIdentifiers),
-        StrField("dataRecord", b'\x00')
+        StrField("dataRecord", b'')
     ]
 
 
@@ -516,6 +578,10 @@ class GMLAN_WDBIPR(Packet):
         XByteEnumField('dataIdentifier', 0, GMLAN_RDBI.dataIdentifiers)
     ]
 
+    def answers(self, other):
+        return isinstance(other, GMLAN_WDBI) \
+            and other.dataIdentifier == self.dataIdentifier
+
 
 bind_layers(GMLAN, GMLAN_WDBIPR, service=0x7B)
 
@@ -608,6 +674,32 @@ bind_layers(GMLAN_RDI, GMLAN_RDI_BC, subfunction=0x82)
 # TODO:This function receive single frame responses... (Implement GMLAN Socket)
 
 
+# ########################DC###################################
+class GMLAN_DC(Packet):
+    name = 'DeviceControl'
+    fields_desc = [
+        XByteField('CPIDNumber', 0),
+        StrFixedLenField('CPIDControlBytes', b"", 5)
+    ]
+
+
+bind_layers(GMLAN, GMLAN_DC, service=0xAE)
+
+
+class GMLAN_DCPR(Packet):
+    name = 'DeviceControlPositiveResponse'
+    fields_desc = [
+        XByteField('CPIDNumber', 0)
+    ]
+
+    def answers(self, other):
+        return isinstance(other, GMLAN_DC) \
+            and other.CPIDNumber == self.CPIDNumber
+
+
+bind_layers(GMLAN, GMLAN_DCPR, service=0xEE)
+
+
 # ########################NRC###################################
 class GMLAN_NR(Packet):
     negativeResponseCodes = {
@@ -633,22 +725,10 @@ class GMLAN_NR(Packet):
         ShortField('deviceControlLimitExceeded', 0)
     ]
 
+    def answers(self, other):
+        return self.requestServiceId == other.service and \
+            (self.returnCode != 0x78 or
+             conf.contribs['GMLAN']['treat-response-pending-as-answer'])
+
 
 bind_layers(GMLAN, GMLAN_NR, service=0x7f)
-
-
-# ##################################################################
-# ######################## UTILS ###################################
-# ##################################################################
-
-
-class GMLAN_TesterPresentSender(PeriodicSenderThread):
-    def __init__(self, sock, pkt=GMLAN(service="TesterPresent"), interval=2):
-        """ Thread to send TesterPresent messages packets periodically
-
-        Args:
-            sock: socket where packet is sent periodically
-            pkt: packet to send
-            interval: interval between two packets
-        """
-        PeriodicSenderThread.__init__(self, sock, pkt, interval)
