@@ -1,5 +1,3 @@
-#! /usr/bin/env python
-#
 # scapy.contrib.description = EtherCat
 # scapy.contrib.status = loads
 
@@ -53,7 +51,7 @@ from scapy.error import log_runtime, Scapy_Exception
 from scapy.fields import BitField, ByteField, LEShortField, FieldListField, \
     LEIntField, FieldLenField, _EnumField, EnumField
 from scapy.layers.l2 import Ether, Dot1Q
-from scapy.modules import six
+import scapy.modules.six as six
 from scapy.packet import bind_layers, Packet, Padding
 
 '''
