@@ -23,17 +23,28 @@ import types
 import warnings
 from random import choice
 
-# Never add any global import, in main.py, that would trigger a warning message  # noqa: E501
-# before the console handlers gets added in interact()
-from scapy.error import log_interactive, log_loading, log_scapy, \
-    Scapy_Exception, ScapyColoredFormatter
+# Never add any global import, in main.py, that would trigger a
+# warning message before the console handlers gets added in interact()
+from scapy.error import (
+    log_interactive,
+    log_loading,
+    Scapy_Exception,
+)
 import scapy.modules.six as six
 from scapy.themes import DefaultTheme, BlackAndWhite, apply_ipython_style
 from scapy.consts import WINDOWS
 
-IGNORED = list(six.moves.builtins.__dict__)
+from scapy.compat import (
+    cast,
+    Any,
+    Dict,
+    List,
+    Optional,
+    Tuple,
+    Union
+)
 
-GLOBKEYS = []
+IGNORED = list(six.moves.builtins.__dict__)
 
 LAYER_ALIASES = {
     "tls": "tls.all"
@@ -44,14 +55,16 @@ QUOTES = [
     ("Craft packets like I craft my beer.", "Jean De Clerck"),
     ("Craft packets before they craft you.", "Socrate"),
     ("Craft me if you can.", "IPv6 layer"),
-    ("To craft a packet, you have to be a packet, and learn how to swim in the "  # noqa: E501
-     "wires and in the waves.", "Jean-Claude Van Damme"),
+    ("To craft a packet, you have to be a packet, and learn how to swim in "
+     "the wires and in the waves.", "Jean-Claude Van Damme"),
     ("We are in France, we say Skappee. OK? Merci.", "Sebastien Chabal"),
-    ("Wanna support scapy? Rate it on sectools! http://sectools.org/tool/scapy/", "Satoshi Nakamoto"),  # noqa: E501
+    ("Wanna support scapy? Star us on GitHub!", "Satoshi Nakamoto"),
+    ("What is dead may never die!", "Python 2"),
 ]
 
 
 def _probe_config_file(cf):
+    # type: (str) -> Union[str, None]
     cf_path = os.path.join(os.path.expanduser("~"), cf)
     try:
         os.stat(cf_path)
@@ -61,26 +74,30 @@ def _probe_config_file(cf):
         return cf_path
 
 
-def _read_config_file(cf, _globals=globals(), _locals=locals(), interactive=True):  # noqa: E501
-    """Read a config file: execute a python file while loading scapy, that may contain  # noqa: E501
-    some pre-configured values.
+def _read_config_file(cf, _globals=globals(), _locals=locals(),
+                      interactive=True):
+    # type: (str, Dict[str, Any], Dict[str, Any], bool) -> None
+    """Read a config file: execute a python file while loading scapy, that
+    may contain some pre-configured values.
 
-    If _globals or _locals are specified, they will be updated with the loaded vars.  # noqa: E501
-    This allows an external program to use the function. Otherwise, vars are only available  # noqa: E501
-    from inside the scapy console.
+    If _globals or _locals are specified, they will be updated with
+    the loaded vars.  This allows an external program to use the
+    function. Otherwise, vars are only available from inside the scapy
+    console.
 
     params:
     - _globals: the globals() vars
     - _locals: the locals() vars
-    - interactive: specified whether or not errors should be printed using the scapy console or  # noqa: E501
-    raised.
+    - interactive: specified whether or not errors should be printed
+    using the scapy console or raised.
 
     ex, content of a config.py file:
         'conf.verb = 42\n'
     Manual loading:
         >>> _read_config_file("./config.py"))
         >>> conf.verb
-        42
+        2
+
     """
     log_loading.debug("Loading config file [%s]", cf)
     try:
@@ -96,10 +113,12 @@ def _read_config_file(cf, _globals=globals(), _locals=locals(), interactive=True
     except Exception:
         if interactive:
             raise
-        log_loading.exception("Error during evaluation of config file [%s]", cf)  # noqa: E501
+        log_loading.exception("Error during evaluation of config file [%s]",
+                              cf)
 
 
 def _validate_local(x):
+    # type: (str) -> bool
     """Returns whether or not a variable should be imported.
     Will return False for any default modules (sys), or if
     they are detected as private vars (starting with a _)"""
@@ -109,10 +128,10 @@ def _validate_local(x):
 
 DEFAULT_PRESTART_FILE = _probe_config_file(".scapy_prestart.py")
 DEFAULT_STARTUP_FILE = _probe_config_file(".scapy_startup.py")
-SESSION = None
 
 
 def _usage():
+    # type: () -> None
     print(
         "Usage: scapy.py [-s sessionfile] [-c new_startup_file] "
         "[-p new_prestart_file] [-C] [-P] [-H]\n"
@@ -130,6 +149,7 @@ def _usage():
 
 
 def _load(module, globals_dict=None, symb_list=None):
+    # type: (str, Optional[Dict[str, Any]], Optional[List[str]]) -> None
     """Loads a Python module to make variables, objects and functions
 available globally.
 
@@ -159,6 +179,7 @@ symbols to the global symbol table.
 
 
 def load_module(name, globals_dict=None, symb_list=None):
+    # type: (str, Optional[Dict[str, Any]], Optional[List[str]]) -> None
     """Loads a Scapy module to make variables, objects and functions
     available globally.
 
@@ -168,6 +189,7 @@ def load_module(name, globals_dict=None, symb_list=None):
 
 
 def load_layer(name, globals_dict=None, symb_list=None):
+    # type: (str, Optional[Dict[str, Any]], Optional[List[str]]) -> None
     """Loads a Scapy layer module to make variables, objects and functions
     available globally.
 
@@ -177,6 +199,7 @@ def load_layer(name, globals_dict=None, symb_list=None):
 
 
 def load_contrib(name, globals_dict=None, symb_list=None):
+    # type: (str, Optional[Dict[str, Any]], Optional[List[str]]) -> None
     """Loads a Scapy contrib module to make variables, objects and
     functions available globally.
 
@@ -197,11 +220,17 @@ def load_contrib(name, globals_dict=None, symb_list=None):
             raise e  # Let's raise the original error to avoid confusion
 
 
-def list_contrib(name=None, ret=False, _debug=False):
+def list_contrib(name=None,  # type: Optional[str]
+                 ret=False,  # type: bool
+                 _debug=False  # type: bool
+                 ):
+    # type: (...) -> Optional[List[Dict[str, str]]]
     """Show the list of all existing contribs.
-    Params:
-     - name: filter to search the contribs
-     - ret: whether the function should return a dict instead of printing it
+
+    :param name: filter to search the contribs
+    :param ret: whether the function should return a dict instead of
+        printing it
+    :returns: None or a dictionary containing the results if ret=True
     """
     # _debug: checks that all contrib modules have correctly defined:
     # # scapy.contrib.description = [...]
@@ -214,7 +243,7 @@ def list_contrib(name=None, ret=False, _debug=False):
         name = "*.py"
     elif "*" not in name and "?" not in name and not name.endswith(".py"):
         name += ".py"
-    results = []
+    results = []  # type: List[Dict[str, str]]
     dir_path = os.path.join(os.path.dirname(__file__), "contrib")
     if sys.version_info >= (3, 5):
         name = os.path.join(dir_path, "**", name)
@@ -228,17 +257,17 @@ def list_contrib(name=None, ret=False, _debug=False):
             continue
         if mod.endswith(".py"):
             mod = mod[:-3]
-        desc = {"description": None, "status": None, "name": mod}
+        desc = {"description": "", "status": "", "name": mod}
         with io.open(f, errors="replace") as fd:
-            for l in fd:
-                if l[0] != "#":
+            for line in fd:
+                if line[0] != "#":
                     continue
-                p = l.find("scapy.contrib.")
+                p = line.find("scapy.contrib.")
                 if p >= 0:
                     p += 14
-                    q = l.find("=", p)
-                    key = l[p:q].strip()
-                    value = l[q + 1:].strip()
+                    q = line.find("=", p)
+                    key = line[p:q].strip()
+                    value = line[q + 1:].strip()
                     desc[key] = value
                 if desc["status"] == "skip":
                     break
@@ -257,6 +286,7 @@ def list_contrib(name=None, ret=False, _debug=False):
     else:
         for desc in results:
             print("%(name)-20s: %(description)-40s status=%(status)s" % desc)
+        return None
 
 
 ##############################
@@ -264,15 +294,17 @@ def list_contrib(name=None, ret=False, _debug=False):
 ##############################
 
 def update_ipython_session(session):
+    # type: (Dict[str, Any]) -> None
     """Updates IPython session with a custom one"""
     try:
-        global get_ipython
+        from IPython import get_ipython
         get_ipython().user_ns.update(session)
     except Exception:
         pass
 
 
-def save_session(fname=None, session=None, pickleProto=-1):
+def save_session(fname="", session=None, pickleProto=-1):
+    # type: (str, Optional[Dict[str, Any]], int) -> None
     """Save current Scapy session to the file specified in the fname arg.
 
     params:
@@ -281,31 +313,34 @@ def save_session(fname=None, session=None, pickleProto=-1):
      - pickleProto: pickle proto version (default: -1 = latest)"""
     from scapy import utils
     from scapy.config import conf, ConfClass
-    if fname is None:
+    if not fname:
         fname = conf.session
         if not fname:
             conf.session = fname = utils.get_temp_file(keep=True)
-    log_interactive.info("Use [%s] as session file" % fname)
+    log_interactive.info("Use [%s] as session file", fname)
 
-    if session is None:
+    if not session:
         try:
+            from IPython import get_ipython
             session = get_ipython().user_ns
         except Exception:
             session = six.moves.builtins.__dict__["scapy_session"]
 
-    to_be_saved = session.copy()
+    to_be_saved = cast(Dict[str, Any], session).copy()
     if "__builtins__" in to_be_saved:
         del(to_be_saved["__builtins__"])
 
     for k in list(to_be_saved):
         i = to_be_saved[k]
-        if hasattr(i, "__module__") and (k[0] == "_" or i.__module__.startswith("IPython")):  # noqa: E501
+        if hasattr(i, "__module__") and (k[0] == "_" or
+                                         i.__module__.startswith("IPython")):
             del(to_be_saved[k])
         if isinstance(i, ConfClass):
             del(to_be_saved[k])
         elif isinstance(i, (type, type, types.ModuleType)):
             if k[0] != "_":
-                log_interactive.error("[%s] (%s) can't be saved.", k, type(to_be_saved[k]))  # noqa: E501
+                log_interactive.error("[%s] (%s) can't be saved.", k,
+                                      type(to_be_saved[k]))
             del(to_be_saved[k])
 
     try:
@@ -316,10 +351,10 @@ def save_session(fname=None, session=None, pickleProto=-1):
     f = gzip.open(fname, "wb")
     six.moves.cPickle.dump(to_be_saved, f, pickleProto)
     f.close()
-    del f
 
 
 def load_session(fname=None):
+    # type: (Optional[Union[str, None]]) -> None
     """Load current Scapy session from the file specified in the fname arg.
     This will erase any existing session.
 
@@ -342,10 +377,11 @@ def load_session(fname=None):
     scapy_session.update(s)
     update_ipython_session(scapy_session)
 
-    log_loading.info("Loaded session [%s]" % fname)
+    log_loading.info("Loaded session [%s]", fname)
 
 
 def update_session(fname=None):
+    # type: (Optional[Union[str, None]]) -> None
     """Update current Scapy session from the file specified in the fname arg.
 
     params:
@@ -362,33 +398,44 @@ def update_session(fname=None):
     update_ipython_session(scapy_session)
 
 
-def init_session(session_name, mydict=None):
+def init_session(session_name,  # type: Optional[Union[str, None]]
+                 mydict=None  # type: Optional[Union[Dict[str, Any], None]]
+                 ):
+    # type: (...) -> Tuple[Dict[str, Any], List[str]]
     from scapy.config import conf
-    global SESSION
-    global GLOBKEYS
+    SESSION = {}  # type: Dict[str, Any]
+    GLOBKEYS = []  # type: List[str]
 
-    scapy_builtins = {k: v for k, v in six.iteritems(importlib.import_module(".all", "scapy").__dict__) if _validate_local(k)}  # noqa: E501
+    scapy_builtins = {k: v
+                      for k, v in six.iteritems(
+                          importlib.import_module(".all", "scapy").__dict__
+                      )
+                      if _validate_local(k)}
     six.moves.builtins.__dict__.update(scapy_builtins)
     GLOBKEYS.extend(scapy_builtins)
     GLOBKEYS.append("scapy_session")
-    scapy_builtins = None
 
     if session_name:
         try:
             os.stat(session_name)
         except OSError:
-            log_loading.info("New session [%s]" % session_name)
+            log_loading.info("New session [%s]", session_name)
         else:
             try:
                 try:
-                    SESSION = six.moves.cPickle.load(gzip.open(session_name, "rb"))  # noqa: E501
+                    SESSION = six.moves.cPickle.load(gzip.open(session_name,
+                                                               "rb"))
                 except IOError:
                     SESSION = six.moves.cPickle.load(open(session_name, "rb"))
-                log_loading.info("Using session [%s]" % session_name)
+                log_loading.info("Using session [%s]", session_name)
+            except ValueError:
+                msg = "Error opening Python3 pickled session on Python2 [%s]"
+                log_loading.error(msg, session_name)
             except EOFError:
-                log_loading.error("Error opening session [%s]" % session_name)
+                log_loading.error("Error opening session [%s]", session_name)
             except AttributeError:
-                log_loading.error("Error opening session [%s]. Attribute missing" % session_name)  # noqa: E501
+                log_loading.error("Error opening session [%s]. "
+                                  "Attribute missing", session_name)
 
         if SESSION:
             if "conf" in SESSION:
@@ -409,37 +456,30 @@ def init_session(session_name, mydict=None):
         six.moves.builtins.__dict__["scapy_session"].update(mydict)
         update_ipython_session(mydict)
         GLOBKEYS.extend(mydict)
+    return SESSION, GLOBKEYS
 
 ################
 #     Main     #
 ################
 
 
-def scapy_delete_temp_files():
-    from scapy.config import conf
-    for f in conf.temp_files:
-        try:
-            os.unlink(f)
-        except Exception:
-            pass
-    del(conf.temp_files[:])
-
-
 def _prepare_quote(quote, author, max_len=78):
+    # type: (str, str, int) -> List[str]
     """This function processes a quote and returns a string that is ready
 to be used in the fancy prompt.
 
     """
-    quote = quote.split(' ')
+    _quote = quote.split(' ')
     max_len -= 6
     lines = []
-    cur_line = []
+    cur_line = []  # type: List[str]
 
     def _len(line):
+        # type: (List[str]) -> int
         return sum(len(elt) for elt in line) + len(line) - 1
-    while quote:
-        if not cur_line or (_len(cur_line) + len(quote[0]) - 1 <= max_len):
-            cur_line.append(quote.pop(0))
+    while _quote:
+        if not cur_line or (_len(cur_line) + len(_quote[0]) - 1 <= max_len):
+            cur_line.append(_quote.pop(0))
             continue
         lines.append('   | %s' % ' '.join(cur_line))
         cur_line = []
@@ -451,37 +491,14 @@ to be used in the fancy prompt.
 
 
 def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
-    """Starts Scapy's console."""
-    global SESSION
-    global GLOBKEYS
-
-    try:
-        if WINDOWS:
-            # colorama is bundled within IPython.
-            # logging.StreamHandler will be overwritten when called,
-            # We can't wait for IPython to call it
-            import colorama
-            colorama.init()
-        # Success
-        console_handler = logging.StreamHandler()
-        console_handler.setFormatter(
-            ScapyColoredFormatter(
-                "%(levelname)s: %(message)s",
-            )
-        )
-    except ImportError:
-        # Failure: ignore colors in the logger
-        console_handler = logging.StreamHandler()
-        console_handler.setFormatter(
-            logging.Formatter(
-                "%(levelname)s: %(message)s",
-            )
-        )
-    log_scapy.addHandler(console_handler)
-
+    # type: (Optional[Any], Optional[Any], Optional[Any], int) -> None
+    """
+    Starts Scapy's console.
+    """
     # We're in interactive mode, let's throw the DeprecationWarnings
     warnings.simplefilter("always")
 
+    # Set interactive mode, load the color scheme
     from scapy.config import conf
     conf.interactive = True
     conf.color_theme = DefaultTheme()
@@ -498,27 +515,30 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
 
     try:
         opts = getopt.getopt(argv[1:], "hs:Cc:Pp:d:H")
-        for opt, parm in opts[0]:
+        for opt, param in opts[0]:
             if opt == "-h":
                 _usage()
             elif opt == "-H":
                 conf.fancy_prompt = False
-                conf.verb = 30
+                conf.verb = 1
+                conf.logLevel = logging.WARNING
             elif opt == "-s":
-                session_name = parm
+                session_name = param
             elif opt == "-c":
-                STARTUP_FILE = parm
+                STARTUP_FILE = param
             elif opt == "-C":
                 STARTUP_FILE = None
             elif opt == "-p":
-                PRESTART_FILE = parm
+                PRESTART_FILE = param
             elif opt == "-P":
                 PRESTART_FILE = None
             elif opt == "-d":
                 conf.logLevel = max(1, conf.logLevel - 10)
 
         if len(opts[1]) > 0:
-            raise getopt.GetoptError("Too many parameters : [%s]" % " ".join(opts[1]))  # noqa: E501
+            raise getopt.GetoptError(
+                "Too many parameters : [%s]" % " ".join(opts[1])
+            )
 
     except getopt.GetoptError as msg:
         log_loading.error(msg)
@@ -527,7 +547,7 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
     # Reset sys.argv, otherwise IPython thinks it is for him
     sys.argv = sys.argv[:1]
 
-    init_session(session_name, mydict)
+    SESSION, GLOBKEYS = init_session(session_name, mydict)
 
     if STARTUP_FILE:
         _read_config_file(STARTUP_FILE, interactive=True)
@@ -616,7 +636,7 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
             quote, author = choice(QUOTES)
             the_banner.extend(_prepare_quote(quote, author, max_len=39))
             the_banner.append("   |")
-        the_banner = "\n".join(
+        banner_text = "\n".join(
             logo + banner for logo, banner in six.moves.zip_longest(
                 (conf.color_theme.logo(line) for line in the_logo),
                 (conf.color_theme.success(line) for line in the_banner),
@@ -624,13 +644,13 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
             )
         )
     else:
-        the_banner = "Welcome to Scapy (%s)" % conf.version
+        banner_text = "Welcome to Scapy (%s)" % conf.version
     if mybanner is not None:
-        the_banner += "\n"
-        the_banner += mybanner
+        banner_text += "\n"
+        banner_text += mybanner
 
     if IPYTHON:
-        banner = the_banner + " using IPython %s\n" % IPython.__version__
+        banner = banner_text + " using IPython %s\n" % IPython.__version__
         try:
             from traitlets.config.loader import Config
         except ImportError:
@@ -645,19 +665,26 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
                     exec_lines=["print(\"\"\"" + banner + "\"\"\")"]
                 )
             except Exception:
-                code.interact(banner=the_banner, local=SESSION)
+                code.interact(banner=banner_text, local=SESSION)
         else:
             cfg = Config()
             try:
-                get_ipython
-            except NameError:
-                # Set "classic" prompt style when launched from run_scapy(.bat) files  # noqa: E501
-                # Register and apply scapy color+prompt style
+                from IPython import get_ipython
+                if not get_ipython():
+                    raise ImportError
+            except ImportError:
+                # Set "classic" prompt style when launched from
+                # run_scapy(.bat) files Register and apply scapy
+                # color+prompt style
                 apply_ipython_style(shell=cfg.TerminalInteractiveShell)
                 cfg.TerminalInteractiveShell.confirm_exit = False
                 cfg.TerminalInteractiveShell.separate_in = u''
             if int(IPython.__version__[0]) >= 6:
-                cfg.TerminalInteractiveShell.term_title_format = "Scapy v%s" % conf.version  # noqa: E501
+                cfg.TerminalInteractiveShell.term_title_format = ("Scapy v%s" %
+                                                                  conf.version)
+                # As of IPython 6-7, the jedi completion module is a dumpster
+                # of fire that should be scrapped never to be seen again.
+                cfg.Completer.use_jedi = False
             else:
                 cfg.TerminalInteractiveShell.term_title = False
             cfg.HistoryAccessor.hist_file = conf.histfile
@@ -666,9 +693,9 @@ def interact(mydict=None, argv=None, mybanner=None, loglevel=logging.INFO):
             try:
                 start_ipython(config=cfg, user_ns=SESSION)
             except (AttributeError, TypeError):
-                code.interact(banner=the_banner, local=SESSION)
+                code.interact(banner=banner_text, local=SESSION)
     else:
-        code.interact(banner=the_banner, local=SESSION)
+        code.interact(banner=banner_text, local=SESSION)
 
     if conf.session:
         save_session(conf.session, SESSION)
