@@ -3,22 +3,24 @@
 # Copyright (C) Philippe Biondi <phil@secdev.org>
 # This program is published under a GPLv2 license
 
-# flake8: noqa: E501
-
 """
 Unit testing infrastructure for Scapy
 """
 
-from __future__ import absolute_import
 from __future__ import print_function
-import sys
+
+import bz2
+import copy
+import code
 import getopt
 import glob
-import importlib
 import hashlib
-import copy
-import bz2
+import importlib
+import json
+import logging
+import os
 import os.path
+import sys
 import time
 import traceback
 import warnings
@@ -27,7 +29,35 @@ import zlib
 from scapy.consts import WINDOWS
 import scapy.modules.six as six
 from scapy.modules.six.moves import range
+from scapy.config import conf
 from scapy.compat import base64_bytes, bytes_hex, plain_str
+from scapy.themes import DefaultTheme, BlackAndWhite
+
+
+# Check UTF-8 support #
+
+def _utf8_support():
+    """
+    Check UTF-8 support for the output
+    """
+    try:
+        if six.PY2:
+            return False
+        if WINDOWS:
+            return (sys.stdout.encoding == "utf-8")
+        return True
+    except AttributeError:
+        return False
+
+
+if _utf8_support():
+    arrow = "\u2514"
+    dash = "\u2501"
+    checkmark = "\u2713"
+else:
+    arrow = "->"
+    dash = "--"
+    checkmark = "OK"
 
 
 #   Util class   #
@@ -35,11 +65,12 @@ from scapy.compat import base64_bytes, bytes_hex, plain_str
 class Bunch:
     __init__ = lambda self, **kw: setattr(self, '__dict__', kw)
 
+
 def retry_test(func):
     """Retries the passed function 3 times before failing"""
     success = False
     ex = Exception("Unknown")
-    for i in six.moves.range(3):
+    for _ in six.moves.range(3):
         try:
             result = func()
         except Exception as e:
@@ -53,6 +84,15 @@ def retry_test(func):
     assert success
     return result
 
+
+def scapy_path(fname):
+    """Resolves a path relative to scapy's root folder"""
+    if fname.startswith('/'):
+        fname = fname[1:]
+    return os.path.abspath(os.path.join(
+        os.path.dirname(__file__), '../../', fname
+    ))
+
 #    Import tool    #
 
 
@@ -151,6 +191,7 @@ class Format(EnumClass):
     HTML = 3
     LATEX = 4
     XUNIT = 5
+    LIVE = 6
 
 
 #    TEST CLASSES    #
@@ -186,11 +227,15 @@ class TestCampaign(TestClass):
         self.preexec_output = None
         self.end_pos = 0
         self.interrupted = False
+        self.duration = 0.0
+
     def add_testset(self, testset):
         self.campaign.append(testset)
         testset.keywords.update(self.keywords)
+
     def trunc(self, index):
         self.campaign = self.campaign[:index]
+
     def startNum(self, beginpos):
         for ts in self:
             for t in ts:
@@ -219,8 +264,10 @@ class TestSet(TestClass):
     def add_test(self, test):
         self.tests.append(test)
         test.keywords.update(self.keywords)
+
     def trunc(self, index):
         self.tests = self.tests[:index]
+
     def __iter__(self):
         return self.tests.__iter__()
 
@@ -230,19 +277,26 @@ class UnitTest(TestClass):
         self.name = name
         self.test = ""
         self.comments = ""
-        self.result = "passed" # make instance True at init to have a different truth value than None
+        self.result = "passed"
+        self.fresult = ""
+        # make instance True at init to have a different truth value than None
+        self.duration = 0
         self.output = ""
         self.num = -1
         self.keywords = set()
         self.crc = None
         self.expand = 1
 
-    def decode(self):
+    def prepare(self, theme):
         if six.PY2:
             self.test = self.test.decode("utf8", "ignore")
             self.output = self.output.decode("utf8", "ignore")
             self.comments = self.comments.decode("utf8", "ignore")
             self.result = self.result.decode("utf8", "ignore")
+        if self.result == "passed":
+            self.fresult = theme.success(self.result)
+        else:
+            self.fresult = theme.fail(self.result)
 
     def __nonzero__(self):
         return self.result == "passed"
@@ -260,8 +314,8 @@ def parse_config_file(config_path, verb=3):
       "onlyfailed": false,
       "verb": 3,
       "dump": 0,
+      "docs": 0,
       "crc": true,
-      "scapy": "scapy",
       "preexec": {},
       "global_preexec": "",
       "outputfile": null,
@@ -274,11 +328,10 @@ def parse_config_file(config_path, verb=3):
     }
 
     """
-    import json
     with open(config_path) as config_file:
-        data = json.load(config_file, encoding="utf8")
+        data = json.load(config_file)
         if verb > 2:
-            print("### Loaded config file", config_path, file=sys.stderr)
+            print(" %s Loaded config file" % arrow, config_path)
 
     def get_if_exist(key, default):
         return data[key] if key in data else default
@@ -288,7 +341,7 @@ def parse_config_file(config_path, verb=3):
                  onlyfailed=get_if_exist("onlyfailed", False),
                  verb=get_if_exist("verb", 3),
                  dump=get_if_exist("dump", 0), crc=get_if_exist("crc", 1),
-                 scapy=get_if_exist("scapy", "scapy"),
+                 docs=get_if_exist("docs", 0),
                  preexec=get_if_exist("preexec", {}),
                  global_preexec=get_if_exist("global_preexec", ""),
                  outfile=get_if_exist("outputfile", sys.stdout),
@@ -309,38 +362,38 @@ def parse_campaign_file(campaign_file):
     test = None
     testnb = 0
 
-    for l in campaign_file.readlines():
-        if l[0] == '#':
+    for line in campaign_file.readlines():
+        if line[0] == '#':
             continue
-        if l[0] == "~":
-            (test or testset or test_campaign).add_keywords(l[1:].split())
-        elif l[0] == "%":
-            test_campaign.title = l[1:].strip()
-        elif l[0] == "+":
-            testset = TestSet(l[1:].strip())
+        if line[0] == "~":
+            (test or testset or test_campaign).add_keywords(line[1:].split())
+        elif line[0] == "%":
+            test_campaign.title = line[1:].strip()
+        elif line[0] == "+":
+            testset = TestSet(line[1:].strip())
             test_campaign.add_testset(testset)
             test = None
-        elif l[0] == "=":
-            test = UnitTest(l[1:].strip())
+        elif line[0] == "=":
+            test = UnitTest(line[1:].strip())
             test.num = testnb
             testnb += 1
             if testset is None:
                 error_m = "Please create a test set (i.e. '+' section)."
                 raise getopt.GetoptError(error_m)
             testset.add_test(test)
-        elif l[0] == "*":
+        elif line[0] == "*":
             if test is not None:
-                test.comments += l[1:]
+                test.comments += line[1:]
             elif testset is not None:
-                testset.comments += l[1:]
+                testset.comments += line[1:]
             else:
-                test_campaign.headcomments += l[1:]
+                test_campaign.headcomments += line[1:]
         else:
             if test is None:
-                if l.strip():
-                    print("Unknown content [%s]" % l.strip(), file=sys.stderr)
+                if line.strip():
+                    raise ValueError("Unknown content [%s]" % line.strip())
             else:
-                test.test += l
+                test.test += line
     return test_campaign
 
 
@@ -370,6 +423,33 @@ def dump_campaign(test_campaign):
                 print("    %s%s" % (c, k))
 
 
+def docs_campaign(test_campaign):
+    print("%(title)s" % test_campaign)
+    print("=" * (len(test_campaign.title)))
+    print()
+    if len(test_campaign.headcomments):
+        print("%s" % test_campaign.headcomments.strip().replace("\n", ""))
+        print()
+    for ts in test_campaign:
+        print("%s" % ts.name)
+        print("-" * len(ts.name))
+        print()
+        if len(ts.comments):
+            print("%s" % ts.comments.strip().replace("\n", ""))
+            print()
+        for t in ts:
+            print("%s" % t.name)
+            print("^" * len(t.name))
+            print()
+            if len(t.comments):
+                print("%s" % t.comments.strip().replace("\n", ""))
+                print()
+            print("Usage example::")
+            for line in t.test.split('\n'):
+                if not line.rstrip().endswith('# no_docs'):
+                    print("\t%s" % line)
+
+
 #    COMPUTE CAMPAIGN DIGESTS    #
 if six.PY2:
     def crc32(x):
@@ -436,10 +516,12 @@ def remove_empty_testsets(test_campaign):
     test_campaign.campaign = [ts for ts in test_campaign.campaign if ts.tests]
 
 
-#### RUN TEST #####
+# RUN TEST #
 
-def run_test(test, get_interactive_session, verb=3, ignore_globals=None, my_globals=None):
+def run_test(test, get_interactive_session, theme, verb=3,
+             ignore_globals=None, my_globals=None):
     """An internal UTScapy function to run a single test"""
+    start_time = time.time()
     test.output, res = get_interactive_session(test.test.strip(), ignore_globals=ignore_globals, verb=verb, my_globals=my_globals)
     test.result = "failed"
     try:
@@ -452,6 +534,7 @@ def run_test(test, get_interactive_session, verb=3, ignore_globals=None, my_glob
         test.output += "UTscapy: Error during result interpretation:\n"
         test.output += "".join(traceback.format_exception(sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2],))
     finally:
+        test.duration = time.time() - start_time
         if test.result == "failed":
             from scapy.sendrecv import debug
             # Add optional debugging data to log
@@ -459,56 +542,100 @@ def run_test(test, get_interactive_session, verb=3, ignore_globals=None, my_glob
                 cls, val = debug.crashed_on
                 test.output += "\n\nPACKET DISSECTION FAILED ON:\n %s(hex_bytes('%s'))" % (cls.__name__, plain_str(bytes_hex(val)))
                 debug.crashed_on = None
-        test.decode()
-        if verb > 1:
-            print("%(result)6s %(crc)s %(name)s" % test, file=sys.stderr)
+        test.prepare(theme)
+        if verb > 2:
+            print("%(fresult)6s %(crc)s %(duration)06.2fs %(name)s" % test)
+        elif verb > 1:
+            print("%(fresult)6s %(crc)s %(name)s" % test)
 
     return bool(test)
 
-#### RUN CAMPAIGN #####
+# RUN CAMPAIGN #
+
 
 def import_UTscapy_tools(ses):
     """Adds UTScapy tools directly to a session"""
-    ses["retry_test"] = retry_test
     ses["Bunch"] = Bunch
+    ses["retry_test"] = retry_test
+    ses["scapy_path"] = scapy_path
+    if WINDOWS:
+        from scapy.arch.windows import _route_add_loopback
+        _route_add_loopback()
+        ses["conf"].ifaces = conf.ifaces
+        ses["conf"].route.routes = conf.route.routes
+        ses["conf"].route6.routes = conf.route6.routes
 
-def run_campaign(test_campaign, get_interactive_session, verb=3, ignore_globals=None):  # noqa: E501
+
+def run_campaign(test_campaign, get_interactive_session, theme,
+                 drop_to_interpreter=False, verb=3,
+                 ignore_globals=None, scapy_ses=None):
     passed = failed = 0
-    scapy_ses = importlib.import_module(".all", "scapy").__dict__
-    import_UTscapy_tools(scapy_ses)
     if test_campaign.preexec:
-        test_campaign.preexec_output = get_interactive_session(test_campaign.preexec.strip(), ignore_globals=ignore_globals, my_globals=scapy_ses)[0]
+        test_campaign.preexec_output = get_interactive_session(
+            test_campaign.preexec.strip(), ignore_globals=ignore_globals,
+            my_globals=scapy_ses)[0]
+
+    # Drop
+    def drop(scapy_ses):
+        code.interact(banner="Test '%s' failed. "
+                             "exit() to stop, Ctrl-D to leave "
+                             "this interpreter and continue "
+                             "with the current test campaign"
+                             % t.name, local=scapy_ses)
+
     try:
         for i, testset in enumerate(test_campaign):
             for j, t in enumerate(testset):
-                if run_test(t, get_interactive_session, verb, my_globals=scapy_ses):
+                if run_test(t, get_interactive_session, theme,
+                            verb=verb, my_globals=scapy_ses):
                     passed += 1
                 else:
                     failed += 1
+                    if drop_to_interpreter:
+                        drop(scapy_ses)
+                test_campaign.duration += t.duration
     except KeyboardInterrupt:
         failed += 1
-        testset.trunc(j+1)
-        test_campaign.trunc(i+1)
+        testset.trunc(j + 1)
+        test_campaign.trunc(i + 1)
         test_campaign.interrupted = True
         if verb:
-            print("Campaign interrupted!", file=sys.stderr)
+            print("Campaign interrupted!")
+            if drop_to_interpreter:
+                drop(scapy_ses)
 
     test_campaign.passed = passed
     test_campaign.failed = failed
-    if verb:
-        print("Campaign CRC=%(crc)s  SHA=%(sha)s" % test_campaign, file=sys.stderr)  # noqa: E501
-        print("PASSED=%i FAILED=%i" % (passed, failed), file=sys.stderr)
+    style = [theme.success, theme.fail][bool(failed)]
+    if verb > 2:
+        print("Campaign CRC=%(crc)s in %(duration)06.2fs SHA=%(sha)s" % test_campaign)
+        print(style("PASSED=%i FAILED=%i" % (passed, failed)))
+    elif verb:
+        print("Campaign CRC=%(crc)s  SHA=%(sha)s" % test_campaign)
+        print(style("PASSED=%i FAILED=%i" % (passed, failed)))
     return failed
 
 
 #    INFO LINES    #
 
-def info_line(test_campaign):
+def info_line(test_campaign, theme):
     filename = test_campaign.filename
+    duration = test_campaign.duration
+    if duration > 10:
+        duration = theme.format(duration, "bg_red+white")
+    elif duration > 5:
+        duration = theme.format(duration, "red")
     if filename is None:
-        return "Run %s by UTscapy" % time.ctime()
+        return "Run at %s by UTscapy in %s" % (
+            time.strftime("%H:%M:%S"),
+            duration
+        )
     else:
-        return "Run %s from [%s] by UTscapy" % (time.ctime(), filename)
+        return "Run at %s from [%s] by UTscapy in %s" % (
+            time.strftime("%H:%M:%S"),
+            filename,
+            duration
+        )
 
 
 def html_info_line(test_campaign):
@@ -521,10 +648,15 @@ def html_info_line(test_campaign):
 
 #    CAMPAIGN TO something    #
 
-def campaign_to_TEXT(test_campaign):
-    output="%(title)s\n" % test_campaign
-    output += "-- "+info_line(test_campaign)+"\n\n"
-    output += "Passed=%(passed)i\nFailed=%(failed)i\n\n%(headcomments)s\n" % test_campaign
+def campaign_to_TEXT(test_campaign, theme):
+    ptheme = [lambda x: x, theme.success][bool(test_campaign.passed)]
+    ftheme = [lambda x: x, theme.fail][bool(test_campaign.failed)]
+
+    output = theme.green("\n%(title)s\n" % test_campaign)
+    output += dash + " " + info_line(test_campaign, theme) + "\n"
+    output += ptheme(" " + arrow + " Passed=%(passed)i\n" % test_campaign)
+    output += ftheme(" " + arrow + " Failed=%(failed)i\n" % test_campaign)
+    output += "%(headcomments)s\n" % test_campaign
 
     for testset in test_campaign:
         if any(t.expand for t in testset):
@@ -536,8 +668,8 @@ def campaign_to_TEXT(test_campaign):
     return output
 
 
-def campaign_to_ANSI(test_campaign):
-    return campaign_to_TEXT(test_campaign)
+def campaign_to_ANSI(test_campaign, theme):
+    return campaign_to_TEXT(test_campaign, theme)
 
 
 def campaign_to_xUNIT(test_campaign):
@@ -563,7 +695,7 @@ def campaign_to_HTML(test_campaign):
 
     if test_campaign.crc is not None and test_campaign.sha is not None:
         output += "CRC=<span class=crc>%(crc)s</span> SHA=<span class=crc>%(sha)s</span><br>" % test_campaign
-    output += "<small><em>"+html_info_line(test_campaign)+"</em></small>"
+    output += "<small><em>" + html_info_line(test_campaign) + "</em></small>"
     output += "".join([
         test_campaign.headcomments,
         "\n<p>",
@@ -685,40 +817,51 @@ def campaign_to_LATEX(test_campaign):
     return output
 
 
-#### USAGE ####
+# USAGE #
 
 def usage():
-    print("""Usage: UTscapy [-m module] [-f {text|ansi|HTML|LaTeX}] [-o output_file]
+    print("""Usage: UTscapy [-m module] [-f {text|ansi|HTML|LaTeX|live}] [-o output_file]
                [-t testfile] [-T testfile] [-k keywords [-k ...]] [-K keywords [-K ...]]
-               [-l] [-b] [-d|-D] [-F] [-q[q]] [-P preexecute_python_code]
-               [-s /path/to/scapy] [-c configfile]
+               [-l] [-b] [-d|-D] [-F] [-q[q]] [-i] [-P preexecute_python_code]
+               [-c configfile]
 -t\t\t: provide test files (can be used many times)
 -T\t\t: if -t is used with *, remove a specific file (can be used many times)
 -l\t\t: generate local .js and .css files
 -F\t\t: expand only failed tests
--b\t\t: stop at first failed campaign
+-b\t\t: don't stop at the first failed campaign
 -d\t\t: dump campaign
 -D\t\t: dump campaign and stop
+-R\t\t: dump campaign as reStructuredText
 -C\t\t: don't calculate CRC and SHA
--s\t\t: path to scapy.py
 -c\t\t: load a .utsc config file
+-i\t\t: drop into Python interpreter if test failed
 -q\t\t: quiet mode
 -qq\t\t: [silent mode]
+-x\t\t: use pyannotate
 -n <testnum>\t: only tests whose numbers are given (eg. 1,3-7,12)
+-N\t\t: force non root
 -m <module>\t: additional module to put in the namespace
 -k <kw1>,<kw2>,...\t: include only tests with one of those keywords (can be used many times)
 -K <kw1>,<kw2>,...\t: remove tests with one of those keywords (can be used many times)
 -P <preexecute_python_code>
-""", file=sys.stderr)
+""")
     raise SystemExit
 
 
 #    MAIN    #
 
-def execute_campaign(TESTFILE, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP,
-                     FORMAT, VERB, ONLYFAILED, CRC, autorun_func, pos_begin=0, ignore_globals=None):  # noqa: E501
+def execute_campaign(TESTFILE, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP, DOCS,
+                     FORMAT, VERB, ONLYFAILED, CRC, INTERPRETER,
+                     autorun_func, theme, pos_begin=0,
+                     ignore_globals=None, scapy_ses=None):  # noqa: E501
     # Parse test file
-    test_campaign = parse_campaign_file(TESTFILE)
+    try:
+        test_campaign = parse_campaign_file(TESTFILE)
+    except ValueError as ex:
+        print(
+            theme.red("Error while parsing '%s': '%s'" % (TESTFILE.name, ex))
+        )
+        sys.exit(1)
 
     # Report parameters
     if PREEXEC:
@@ -743,9 +886,20 @@ def execute_campaign(TESTFILE, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP,
         if DUMP > 1:
             sys.exit()
 
+    # Dump campaign as reStructuredText
+    if DOCS:
+        docs_campaign(test_campaign)
+        sys.exit()
+
     # Run tests
     test_campaign.output_file = OUTPUTFILE
-    result = run_campaign(test_campaign, autorun_func[FORMAT], verb=VERB, ignore_globals=None)  # noqa: E501
+    result = run_campaign(
+        test_campaign, autorun_func[FORMAT], theme,
+        drop_to_interpreter=INTERPRETER,
+        verb=VERB,
+        ignore_globals=None,
+        scapy_ses=scapy_ses
+    )
 
     # Shrink passed
     if ONLYFAILED:
@@ -757,9 +911,9 @@ def execute_campaign(TESTFILE, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP,
 
     # Generate report
     if FORMAT == Format.TEXT:
-        output = campaign_to_TEXT(test_campaign)
+        output = campaign_to_TEXT(test_campaign, theme)
     elif FORMAT == Format.ANSI:
-        output = campaign_to_ANSI(test_campaign)
+        output = campaign_to_ANSI(test_campaign, theme)
     elif FORMAT == Format.HTML:
         test_campaign.startNum(pos_begin)
         output = campaign_to_HTML(test_campaign)
@@ -767,6 +921,8 @@ def execute_campaign(TESTFILE, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP,
         output = campaign_to_LATEX(test_campaign)
     elif FORMAT == Format.XUNIT:
         output = campaign_to_xUNIT(test_campaign)
+    elif FORMAT == Format.LIVE:
+        output = ""
 
     return output, (result == 0), test_campaign
 
@@ -775,51 +931,65 @@ def resolve_testfiles(TESTFILES):
     for tfile in TESTFILES[:]:
         if "*" in tfile:
             TESTFILES.remove(tfile)
-            TESTFILES.extend(glob.glob(tfile))
+            TESTFILES.extend(sorted(glob.glob(tfile)))
     return TESTFILES
 
 
 def main():
     argv = sys.argv[1:]
+    logger = logging.getLogger("scapy")
+    logger.addHandler(logging.StreamHandler())
     ignore_globals = list(six.moves.builtins.__dict__)
 
+    import scapy
+    print(dash + " UTScapy - Scapy %s - %s" % (
+        scapy.__version__, sys.version.split(" ")[0]
+    ))
+
     # Parse arguments
 
     FORMAT = Format.ANSI
     OUTPUTFILE = sys.stdout
     LOCAL = 0
     NUM = None
+    NON_ROOT = False
     KW_OK = []
     KW_KO = []
     DUMP = 0
+    DOCS = 0
     CRC = True
     BREAKFAILED = True
     ONLYFAILED = False
     VERB = 3
     GLOB_PREEXEC = ""
     PREEXEC_DICT = {}
-    SCAPY = "scapy"
     MODULES = []
     TESTFILES = []
+    ANNOTATIONS_MODE = False
+    INTERPRETER = False
     try:
-        opts = getopt.getopt(argv, "o:t:T:c:f:hbln:m:k:K:DdCFqP:s:")
+        opts = getopt.getopt(argv, "o:t:T:c:f:hbln:m:k:K:DRdCiFqNP:s:x")
         for opt, optarg in opts[0]:
             if opt == "-h":
                 usage()
             elif opt == "-b":
-                BREAKFAILED =  True
+                BREAKFAILED = False
             elif opt == "-F":
                 ONLYFAILED = True
             elif opt == "-q":
                 VERB -= 1
             elif opt == "-D":
                 DUMP = 2
+            elif opt == "-R":
+                DOCS = 1
             elif opt == "-d":
                 DUMP = 1
             elif opt == "-C":
                 CRC = False
-            elif opt == "-s":
-                SCAPY = optarg
+            elif opt == "-i":
+                INTERPRETER = True
+            elif opt == "-x":
+                ANNOTATIONS_MODE = True
             elif opt == "-P":
                 GLOB_PREEXEC += "\n" + optarg
             elif opt == "-f":
@@ -839,7 +1009,6 @@ def main():
                 VERB = data.verb
                 DUMP = data.dump
                 CRC = data.crc
-                SCAPY = data.scapy
                 PREEXEC_DICT = data.preexec
                 GLOB_PREEXEC = data.global_preexec
                 OUTPUTFILE = data.outfile
@@ -874,6 +1043,8 @@ def main():
                     except ValueError:
                         v1, v2 = [int(e) for e in v.split('-', 1)]
                         NUM.extend(range(v1, v2 + 1))
+            elif opt == "-N":
+                NON_ROOT = True
             elif opt == "-m":
                 MODULES.append(optarg)
             elif opt == "-k":
@@ -881,47 +1052,78 @@ def main():
             elif opt == "-K":
                 KW_KO.extend(optarg.split(","))
 
-        # Discard Python3 tests when using Python2
-        if six.PY2:
-            KW_KO.append("python3_only")
-
-        if VERB > 2:
-            print("### Booting scapy...", file=sys.stderr)
-        try:
-            from scapy import all as scapy
-        except Exception as e:
-            print("[CRITICAL]: Cannot import Scapy: %s" % e, file=sys.stderr)
-            traceback.print_exc()
-            sys.exit(1)  # Abort the tests
-
-        for m in MODULES:
-            try:
-                mod = import_module(m)
-                six.moves.builtins.__dict__.update(mod.__dict__)
-            except ImportError as e:
-                raise getopt.GetoptError("cannot import [%s]: %s" % (m, e))
-
-        if WINDOWS:
-            from scapy.arch.windows import route_add_loopback
-            route_add_loopback()
-
-        # Add SCAPY_ROOT_DIR environment variable, used for tests
-        os.environ['SCAPY_ROOT_DIR'] = os.environ.get("PWD", os.getcwd())
-
     except getopt.GetoptError as msg:
-        print("ERROR:", msg, file=sys.stderr)
+        print("ERROR:", msg)
         raise SystemExit
 
+    if FORMAT in [Format.LIVE, Format.ANSI]:
+        theme = DefaultTheme()
+    else:
+        theme = BlackAndWhite()
+
+    # Disable tests if needed
+
+    # Discard Python3 tests when using Python2
+    if six.PY2:
+        KW_KO.append("python3_only")
+        if VERB > 2:
+            print(" " + arrow + " Python 2 mode")
+    try:
+        if NON_ROOT or os.getuid() != 0:  # Non root
+            # Discard root tests
+            KW_KO.append("netaccess")
+            KW_KO.append("needs_root")
+            if VERB > 2:
+                print(" " + arrow + " Non-root mode")
+    except AttributeError:
+        pass
+
+    if conf.use_pcap:
+        KW_KO.append("not_pcapdnet")
+        if VERB > 2:
+            print(" " + arrow + " libpcap mode")
+
+    KW_KO.append("disabled")
+
+    # Process extras
+    if six.PY3:
+        KW_KO.append("FIXME_py3")
+
+    if ANNOTATIONS_MODE:
+        try:
+            from pyannotate_runtime import collect_types
+        except ImportError:
+            raise ImportError("Please install pyannotate !")
+        collect_types.init_types_collection()
+        collect_types.start()
+
+    if VERB > 2:
+        print(" " + arrow + " Booting scapy...")
+    try:
+        from scapy import all as scapy
+    except Exception as e:
+        print("[CRITICAL]: Cannot import Scapy: %s" % e)
+        traceback.print_exc()
+        sys.exit(1)  # Abort the tests
+
+    for m in MODULES:
+        try:
+            mod = import_module(m)
+            six.moves.builtins.__dict__.update(mod.__dict__)
+        except ImportError as e:
+            raise getopt.GetoptError("cannot import [%s]: %s" % (m, e))
+
     autorun_func = {
         Format.TEXT: scapy.autorun_get_text_interactive_session,
         Format.ANSI: scapy.autorun_get_ansi_interactive_session,
         Format.HTML: scapy.autorun_get_html_interactive_session,
         Format.LATEX: scapy.autorun_get_latex_interactive_session,
         Format.XUNIT: scapy.autorun_get_text_interactive_session,
+        Format.LIVE: scapy.autorun_get_live_interactive_session,
     }
 
     if VERB > 2:
-        print("### Starting tests...", file=sys.stderr)
+        print(" " + arrow + " Discovering tests files...")
 
     glob_output = ""
     glob_result = 0
@@ -941,17 +1143,24 @@ def main():
     pos_begin = 0
 
     runned_campaigns = []
+
+    scapy_ses = importlib.import_module(".all", "scapy").__dict__
+    import_UTscapy_tools(scapy_ses)
+
     # Execute all files
     for TESTFILE in TESTFILES:
         if VERB > 2:
-            print("### Loading:", TESTFILE, file=sys.stderr)
+            print(theme.green(dash + " Loading: %s" % TESTFILE))
         PREEXEC = PREEXEC_DICT[TESTFILE] if TESTFILE in PREEXEC_DICT else GLOB_PREEXEC
         with open(TESTFILE) as testfile:
-            output, result, campaign = execute_campaign(testfile, OUTPUTFILE,
-                                                        PREEXEC, NUM, KW_OK, KW_KO,
-                                                        DUMP, FORMAT, VERB, ONLYFAILED,
-                                                        CRC, autorun_func, pos_begin,
-                                                        ignore_globals)
+            output, result, campaign = execute_campaign(
+                testfile, OUTPUTFILE, PREEXEC, NUM, KW_OK, KW_KO, DUMP, DOCS,
+                FORMAT, VERB, ONLYFAILED, CRC, INTERPRETER,
+                autorun_func, theme,
+                pos_begin=pos_begin,
+                ignore_globals=ignore_globals,
+                scapy_ses=copy.copy(scapy_ses)
+            )
         runned_campaigns.append(campaign)
         pos_begin = campaign.end_pos
         if UNIQUE:
@@ -963,7 +1172,14 @@ def main():
                 break
 
     if VERB > 2:
-        print("### Writing output...", file=sys.stderr)
+        print(
+            checkmark + " All campaigns executed. Writing output..."
+        )
+
+    if ANNOTATIONS_MODE:
+        collect_types.stop()
+        collect_types.dump_stats("pyannotate_results")
+
     # Concenate outputs
     if FORMAT == Format.HTML:
         glob_output = pack_html_campaigns(runned_campaigns, glob_output, LOCAL, glob_title)
@@ -972,15 +1188,25 @@ def main():
     # Note: on Python 2, we force-encode to ignore ascii errors
     # on Python 3, we need to detect the type of stream
     if OUTPUTFILE == sys.stdout:
-        OUTPUTFILE.write(glob_output.encode("utf8", "ignore")
-                         if 'b' in OUTPUTFILE.mode or six.PY2 else glob_output)
+        print(glob_output, file=OUTPUTFILE)
     else:
         with open(OUTPUTFILE, "wb") as f:
             f.write(glob_output.encode("utf8", "ignore")
                     if 'b' in f.mode or six.PY2 else glob_output)
 
-    # Delete scapy's test environment vars
-    del os.environ['SCAPY_ROOT_DIR']
+    # Print end message
+    if VERB > 2:
+        if glob_result == 0:
+            print(theme.green("UTscapy ended successfully"))
+        else:
+            print(theme.red("UTscapy ended with error code %s" % glob_result))
+
+    # Check active threads
+    if VERB > 2:
+        import threading
+        if threading.active_count() > 1:
+            print("\nWARNING: UNFINISHED THREADS")
+            print(threading.enumerate())
 
     # Return state
     return glob_result
